/*
    This file is part of the KDE system
    Copyright (C)  1999,2000 Boloni Laszlo

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

    gofai.cpp - Good Old Fashioned Artificial Intelligence With a Modern Twist
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <kfiledialog.h>
#include <qkeycode.h>
#include <kiconloader.h>
#include <kaccel.h>
#include <kglobal.h>
#include <klocale.h>
#include <kprocess.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <kstdaction.h>
#include <kaction.h>
#include <kimageio.h>
#include <ktempfile.h>
#include <kio/netaccess.h>
//#include <qdom.h>
//#include <qbuffer.h>
//#include <qtextstream.h>

#include "gofai.h"

Gofai::Gofai() :
  KMainWindow(0, 0) {
    KImageIO::registerFormats();
    gofaiConfig::getInstance()->loadConfig(kapp->config());
    initialize(this);
    setCentralWidget(splitter);
    setupActions();
    // initial values
    QWidget *d = QApplication::desktop();

    if (gofaiConfig::getInstance()->detailWindowVisible) {
      int h=d->height();                  // returns screen height
       showDetailWindow();
       resize(500,h-50);
       move(0,0);
    } else {
      hideDetailWindow();
      resize(500,500);
    }

    QTimer::singleShot( 1000, this, SLOT(slotLoadPrevious()) );
    // initializing auto save
    autosaveTimer = new QTimer(this);
    connect( autosaveTimer, SIGNAL(timeout()),
             this, SLOT(slotSaveFile()) );
    restartAutoSave();
}

/**
   Setup actions
 */
void Gofai::setupActions() {
    KStdAction::openNew(this, SLOT(slotNew()), actionCollection());
    KStdAction::open(this, SLOT(slotLoadFile()),actionCollection());
    KStdAction::save(this, SLOT(slotSaveFile()),actionCollection());
    KStdAction::saveAs(this, SLOT(slotSaveFileAs()),actionCollection());
    KStdAction::quit(this, SLOT(close()),actionCollection());
    recentFiles = KStdAction::openRecent(this,
                                         SLOT(slotOpenRecent(const KURL&)),
                                         actionCollection());
    recentFiles->loadEntries(kapp->config());
    KStdAction::find(gofw, SLOT(slotFind()),actionCollection());
    KStdAction::preferences(this, SLOT(slotOptions()),actionCollection());
    new KAction("&History",
                QIconSet(BarIcon("down")),
                0,
                gofw,
                SLOT(slotHistory()),
                actionCollection(),
                "history");
    new KAction("&Random walk",
                QIconSet(BarIcon("randomwalk")),
                0,
                gofw,
                SLOT(slotRandomWalk()),
                actionCollection(),
                "randomwalk");
    new KAction("&Statistics",
                0,
                this,
                SLOT(slotStatistics()),
                actionCollection(),
                "statistics");
    new KAction("Set &mood",
                0,
                this,
                SLOT(slotSetMood()),
                actionCollection(),
                "mood");

    // toggle menubar
    showMenubar = KStdAction::showMenubar(NULL, NULL,
                                          actionCollection());
    showMenubar->setChecked(gofaiConfig::getInstance()->menubarVisible);
    connect(showMenubar,SIGNAL(toggled(bool)),
            this,SLOT(slotToggleMenubar(bool)));


    // toggle toolbar
    showToolbar = KStdAction::showToolbar(NULL, NULL,
                                          actionCollection());
    showToolbar->setChecked(gofaiConfig::getInstance()->toolbarVisible);
    connect(showToolbar,SIGNAL(toggled(bool)),
            this,SLOT(slotToggleToolbar(bool)));

    // toggle statusbar
    showStatusbar = KStdAction::showStatusbar(NULL, NULL,
                                              actionCollection());
    connect(showStatusbar,SIGNAL(toggled(bool)),
            this,SLOT(slotToggleStatusbar(bool)));

    // toggle detail window
    KToggleAction *showDetailWindow =
        new KToggleAction(i18n("&Show details"),
                0,
                0,
                0,
                NULL,
                actionCollection(),
                "toggledetail");
    showDetailWindow->setChecked(gofaiConfig::getInstance()->detailWindowVisible);
    connect(showDetailWindow, SIGNAL(toggled(bool)),
            this, SLOT(slotToggleDetailWindow(bool)));


    new KAction(i18n("Edit &ontologies"),
                0,
                this,
                SLOT(slotEditOntologies()),
                actionCollection(),
                "editOntologies");
    new KAction(i18n("Edit &content types"),
                0,
                this,
                SLOT(slotEditContentTypes()),
                actionCollection(),
                "editContentTypes");
    new KAction(i18n("Edit &link types"),
                0,
                this,
                SLOT(slotEditLinkTypes()),
                actionCollection(),
                "editLinkTypes");
    createGUI("gofai.rc");
}


void Gofai::setupStatusBar() {
    statusBar();
    // a bit of a hack to get it right
    gofaiConfig::getInstance()->statusbarVisible=
      !gofaiConfig::getInstance()->statusbarVisible;
    slotToggleStatusbar(gofaiConfig::getInstance()->statusbarVisible);
}



/**
   Load the previously opened file
 */
void Gofai::slotLoadPrevious() {
    QString loadfile = gofaiConfig::getInstance()->defaultfile;
    if (!loadFile(loadfile)) {
      kdDebug() << "Problem with loading --- creating new!!!\n";
      KMessageBox::information(this,QString("Unable to load previous file: %1.\nStarting new.").arg(loadfile));
    }
}

void Gofai::restartAutoSave() {
    // stop the other one???
    if (gofaiConfig::getInstance()->autosave)
        autosaveTimer->start(gofaiConfig::getInstance()->autosaveInterval,
                             TRUE);
}


/**
   The goal of this is to refresh the transparent
   background
 */
void Gofai::moveEvent (QMoveEvent*) {
  gofw->repaint(false);
}


void Gofai::showDetailWindow() {
    resize(width(), (int) (1.25 * height()));
    QValueList<int> splitinit;
    splitinit.append(80);
    splitinit.append(20);
    splitter->setSizes(splitinit);
    detail->show();
    menuBar()->changeItem(i18n("Hide detail window"), detailMenuItem);
}

void Gofai::hideDetailWindow() {
    resize(width(), (int) (0.8 * height()));
    QValueList<int> splitinit;
    splitinit.append(100);
    splitinit.append(0);
    splitter->setSizes(splitinit);
    detail->hide();
    menuBar()->changeItem(i18n("Show detail window"), detailMenuItem);
}


/**
   Loads the current data source and data view from a
   combined XML file.
 */
bool Gofai::loadFile(const QString & filename) {
  if (! gofai_common::loadFile(filename)) {
      return false;
  };
  gofaiConfig::getInstance()->defaultfile = filename;
  KURL url(filename);
  recentFiles->addURL(url);
  recentFiles->saveEntries(kapp->config());
  refresh();
  return true;
}


/**
   Saves the current data source and
   data view to a combined XML file
 */
bool Gofai::saveToFile(const QString &destination) {
    // Guarantee that we save the current changes
    gofaiConfig::getInstance()->defaultfile = destination;
    bool success = gofai_common::saveToFile(destination);
    refresh();
    if (!success) {
        KMessageBox::sorry( NULL, "Cannot open the destination file");
    }
    return success;
}

/**
   Refreshes the components, caption, detail window
 */
void Gofai::refresh() {
    setCaption( gofaiConfig::getInstance()->defaultfile);
    gofw->refresh();
    detail->setValues(dv->center->val);
}


/**
   Slots
 */

void Gofai::slotNew(){
  int val = KMessageBox::warningYesNo(this,
              "Really create new?\nThis will erase your goph.");
 if (val == KMessageBox::No)
     return;
 dv->newGoph();
 gofaiConfig::getInstance()->defaultfile = "";
 refresh();
}

/**
   Loading a file
 */
void Gofai::slotLoadFile(){
  KURL url = KFileDialog::getOpenURL(QString::null,
                   "*.goph|GOFAI files (*.goph)\n*|All files");
  if (!url.isLocalFile()) {
      // remote file
      gofaiConfig::getInstance()->currentFileIsRemote = true;
      gofaiConfig::getInstance()->currentRemoteURL = url;
      QString tmpFile;
      if( KIO::NetAccess::download(url, tmpFile ) ) {
          loadFile( tmpFile );
          KIO::NetAccess::removeTempFile( tmpFile );
      }
  } else {
      // local file
      QString s = url.path();
      if (!s.isEmpty()){
          loadFile(s);
      }
  }
}

void Gofai::slotOpenRecent(const KURL& url) {
    loadFile(url.path());
}

/**
   FIXME: this is not going to work for the remote URLs.
 */
void Gofai::slotSaveFile(){
  QString s = gofaiConfig::getInstance()->defaultfile;
  if (!s.isEmpty()){
      saveToFile(s);
  } else {
      slotSaveFileAs();
  }
}

void Gofai::slotSaveFileAs(){
  KURL url = KFileDialog::getSaveURL(QString::null,
         "*.goph|GOFAI files (*.goph)\n*|All files");
  // set the recent files dialog
  recentFiles->addURL(url);
  recentFiles->saveEntries(kapp->config());
  if (!url.isLocalFile()) {
      gofaiConfig::getInstance()->currentFileIsRemote = true;
      gofaiConfig::getInstance()->currentRemoteURL = url;
      // create a temporary file and set it
      // to be deleted after this block
      KTempFile tmpFile;
      tmpFile.setAutoDelete(true);
      // now save to the temporary file
      if (!saveToFile(tmpFile.name())) {
          KMessageBox::sorry(this,
                  "An error happened while saving the file. ");
          return;
      }
      // and upload the temporary to the network location
      if (KIO::NetAccess::upload(tmpFile.name(),
              gofaiConfig::getInstance()->currentRemoteURL ) == false) {
          KMessageBox::error(this, KIO::NetAccess::lastErrorString());
          return;
      }
      return;
  } else {
      QString s = url.path();
      if (!s.isEmpty()){
          saveToFile(s);
      }  // else is no filename, user hit "Cancel" or pressed "Escape"
  }
  restartAutoSave();
}


/**
   Calls the configuration dialog
 */
void Gofai::slotOptions() {
  KProcess *proc = new KProcess;
  gofaiConfig::getInstance()->saveConfig(kapp->config());
  *proc << "kcmshell" << "Applications/gofai";
  connect(proc, SIGNAL(processExited(KProcess *)),
          this, SLOT(slotOptionsTerminated(KProcess *)));
  proc->start(KProcess::NotifyOnExit, KProcess::All);
}

/**
   This slot is called when the external configuration
   panel is called.
 */
void Gofai::slotOptionsTerminated(KProcess *proc) {
  // delete the proc
  delete proc;
  gofaiConfig::getInstance()->loadConfig(kapp->config());
  refresh();
}

void Gofai::slotEditOntologies() {
    gofaiEditOntologyDialog eod(this);
    eod.exec();
}

void Gofai::slotEditContentTypes() {
    gofaiEditContentTypesDialog ect(this);
    ect.exec();
}

void Gofai::slotEditLinkTypes() {
    gofaiEditLinkTypesDialog elt(this);
    elt.exec();
}

void Gofai::slotStatistics() {
    gofaiStatisticsWindow sw(this);
    sw.exec();
}

/**
   Toggles the menubar on/off
    todo
 */
void Gofai::slotToggleMenubar(bool value) {
    gofaiConfig::getInstance()->menubarVisible = value;
    if (value) {
        menuBar()->show();
    } else {
        menuBar()->hide();
    }
}

/**
   Toggles the toolbar on/off
 */
void Gofai::slotToggleToolbar(bool value) {
    gofaiConfig::getInstance()->toolbarVisible = value;
    if (value) {
        toolBar()->show();
    } else {
        toolBar()->hide();
    }
}

/**
   Toggles the statusbar on/off
 */
void Gofai::slotToggleStatusbar(bool value) {
    gofaiConfig::getInstance()->toolbarVisible = value;
    if (value) {
        statusBar()->show();
    } else {
        statusBar()->hide();
    }

}

/**
   Toggles the detail window on/off
 */
void Gofai::slotToggleDetailWindow(bool value) {
    gofaiConfig::getInstance()->detailWindowVisible = value;
    if (value) {
        showDetailWindow();
    } else {
        hideDetailWindow();
    }
}

void Gofai::slotSetMood() {
    kdDebug() << "slotSetMood" << endl;
}

void Gofai::slotAutoHide() {
    // currently not implemented
    /*
    resize(300,300);
    enableToolBar(KToolBar::Hide);
    enableStatusBar(KStatusBar::Hide);
    */
}

// this slot is invoked when the last window is going to be destroyed
bool Gofai::queryClose(){
// TODO: annoying query message configurable
    /*
    gofaiConfig::getInstance()->saveConfig(kapp->config());
    if (!dv->isChanged()) {
        return true;
    }
    int val = KMessageBox::warningYesNoCancel(this,
        "Do you want to save the changes to this document?",
        "Leaving GOFAI",
        "Save and exit",
        "Exit without saving");
    switch(val) {
    case KMessageBox::Yes:
      slotSaveFile();
      return true;
    case KMessageBox::No:
      return true;
    default:
      return false;
    };
    */
    slotSaveFile();
    return true;
}

#include "gofai.moc"
