/*
    This file is part of the KDE system
    Copyright (C)  1999 Boloni Laszlo

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.


    gofaiDataView.h

   This allows a limited view of the whole data structure

*/

#include <fstream.h>
#include <math.h>
#include <kdebug.h>
#include <stdlib.h>
#include "gofaiDataView.h"
#include "gofaiVisualArrangement.h"
#include "gofaiConfig.h"
#include "gofaiDataQuery.h"
#include "gofaiDataSource_Memory.h"

/**
   Constructor.
   BTW: this would let me to use different query algorithms.
 */
gofaiDataView::gofaiDataView() {
    count = 0;
    context_count = 0;
    center = NULL;
    dq = new gofaiDataQuery(this);
    ds = new gofaiDataSource_Memory();
};

/**
   Refresh the view, using the query
 */
void gofaiDataView::getView() {
    dq->getView();
}

/**
   Data manipulation stuff
 */

void gofaiDataView::newGoph() {
    delete ds;
    ds = new gofaiDataSource_Memory();
    center = NULL;
    getView();
}

bool gofaiDataView::isChanged() {
    return ds->changed;
}

void gofaiDataView::addItem(gofaiItem *it) {
    ds->addItem(it);
    ds->changed = true;
};

void gofaiDataView::removeItem(gofaiItem *it) {
    ds->removeItem(it);
    ds->changed = true;
};

void gofaiDataView::updateItem(gofaiItem *it) {
    ds->updateItem(it);
    ds->changed = true;
};

gofaiItem *gofaiDataView::getItemByName(QString name) {
    return ds->lookupItemByName(name);
}

gofaiItem *gofaiDataView::getItemById(gofaiID id) {
    return ds->lookupItem(id);
}

void gofaiDataView::getMatchingItems(QString prefix,
                                     QList<gofaiItem> &list) {
    ds->find(prefix, list);
}

/**
   Arranging the visuals in concentric circles around the center,
   based on the level.
 */
int gofaiDataView::createVisuals(int centerx, int centery, int r) {
    gofaiConcentricArrangement arr;
    arr.arrange(*this, centerx, centery, r, r);
    return 0; // FIXME
}

gofaiVisualItem* gofaiDataView::containsItem(gofaiItem *it, bool context) {
    for (int i = 0; i!=count; i++) {
        if (vds[i]->val == it) {
            return vds[i];
        }
    }
    if (context) {
        for (int i = 0; i!=context_count; i++) {
            if (context_vds[i]->val == it) {
                return context_vds[i];
            }
        }
    }
    return NULL;
}

gofaiVisualItem* gofaiDataView::containsItem(gofaiID id, bool context) {
    for (int i = 0; i!=count; i++) {
        if (vds[i]->val->id == id) {
            return vds[i];
        }
    }
    if (context) {
        for (int i = 0; i!=context_count; i++) {
            if (context_vds[i]->val->id == id) {
                return context_vds[i];
            }
        }
    }
    return NULL;
}

/**
   Returns the pointer to the item at the specific position
 */
void gofaiDataView::getItemFromHere(const QPoint &pos,
                                    gofaiDataViewPointer &result,
                                    int type) {
    for(int i=0; i!=count; i++) {
        if (vds[i]->contains(pos, *this) && vds[i]->val->type==type) {
          return result.setTo(i, gofaiDataViewPointer::Main);
        }
    }
    for(int i=0; i!=context_count; i++) {
        if (context_vds[i]->contains(pos, *this)
            && vds[i]->val->type==type) {
          return result.setTo(i, gofaiDataViewPointer::Context);
        }
    }
    return result.invalidate();
}

void gofaiDataView::getNodeFromHere(const QPoint &pos,
                                    gofaiDataViewPointer &result) {
  return getItemFromHere(pos, result, Node);
}

void gofaiDataView::getLinkFromHere(const QPoint &pos,
                                    gofaiDataViewPointer &result) {
  return getItemFromHere(pos, result, Link);
}

void gofaiDataView::setCenter(gofaiVisualItem* vit) {
    center = vit;
    if (history.count() == 0) {
        history.insert(0,vit->val);
    } else if (history.at(0) != vit->val) {
        history.insert(0,vit->val);
    }
}


/**
   Sets the shadow and thicker lines
 */
void gofaiDataView::setActive(gofaiItem* it) {
  // FIXME: for the context, too
  for(int i = 0; i!=count ; i++) {
    gofaiVisualItem *vi = vds[i];
    if (vi->val->type == Link) {
      if (vi->val->from == it->id ||
          vi->val->to == it->id) {
        vi->level = 0;
      } else {
        vi->level = 1;
      }
    }
    if (vi->val->type == Node) {
      if (vi->val==it) {
        vi->shadow = true;
      } else {
        vi->shadow = false;
      }
    }
  }
}



int gofaiDataView::countConnections(gofaiVisualItem *vit) {
    int cnt=0;
    for(int i=0; i!=count; i++) {
        if ((vds[i]->val->type == Link) &&
            ((vds[i]->val->from == vit->val->id) ||
             (vds[i]->val->to == vit->val->id)))
        {
            cnt++;
        }
    }
    return cnt;
}

void gofaiDataView::saveOldToTemp() {
    localviews = count;
    for(int i=0; i!=count; i++) {
        lvds[i] = vds[i];
        vds[i]=NULL;
    }
    count = 0;
};

void gofaiDataView::removeTemp() {
    // and now delete all the things
    for(int i=0; i!=localviews; i++) {
        if (lvds[i] != NULL) {
            delete lvds[i];
        }
    }
    localviews = 0;
};


void gofaiDataView::setNewPosition() {
      for(int i=0; i!=count; i++) {
          if (vds[i]->val->type == Node) {
              vds[i]->setNewPosition();
          }
      }
      for(int i=0; i!=context_count; i++) {
          if (context_vds[i]->val->type == Node) {
              context_vds[i]->setNewPosition();
          }
      }
}

bool gofaiDataView::addToView(gofaiItem *it, int level) {
    /*
      addToView should be smart enough that if there is no space,
      to keep the stronger stuff there
     */
    if (it == NULL) {
        cerr << "Adding null to view!!!\n";
        return false;
    }
    if (it->type ==Node &&
        count >= gofaiConfig::getInstance()->maxNodesInView-1) {
        // kdDebug() << "WARNING: Cannot add another node!, set to max specified" << endl;
      return false;
    }
    if (count >= MAX_ITEMS_IN_VIEW) {
      kdDebug() << "WARNING: Cannot add another item, no more space!" << endl;
      return false;
    }

    bool context = false;
    gofaiVisualItem *vt;
    if ( (vt=containsItem(it)) ) {
        if (vt->level > level)
            vt->level = level;
        return true;
    }
    // if it is in the context, do not add it
    for(int i=0; i!=context_count; i++) {
        if (context_vds[i]->val == it) {
            // cout << it->name<< "It is in the context do not add it\n";
            context = true;
            break;
        }
    }
    for(int i=0; i!=localviews; i++) {
        if ((lvds[i]!=NULL) && (lvds[i]->val == it)) {
            // prevent removing if in the context
            if (context) {
                lvds[i] = NULL;
                kdDebug() << it->name << "We are not adding it: is in the context\n";
                return true;
            }
            // if it was already in the view, keep it.
            vds[count] = lvds[i];
            vds[count]->level = level;
            count++;
            lvds[i] = NULL;
            return true;
        }
    }
    if (context) return true;
    vds[count] = new gofaiVisualItem();
    vds[count]->val = it;
    switch(it->type) {
    case Node:
        vds[count]->type =
            new gofaiVisualItemType_Rectangle(vds[count]);
        break;
    case Link:
        vds[count]->type =
            new gofaiVisualItemType_Link(vds[count]);
    }
    vds[count]->level = level;
    count++;
    return true;
}


bool gofaiDataView::addToContext(gofaiVisualItem *vit) {
  if (context_count < MAX_CONTEXT_ITEMS_IN_VIEW) {
    context_vds[context_count] = vit;
    context_count++;
    return true;
  }
  return false;
}

/**
   Draws or animates. Returns if it is finished
 */
bool gofaiDataView::drawIt(QPainter *p) {
  bool finished = true, val;
  if (gofaiConfig::getInstance()->animation == false) {
    // set the new values here
    setNewPosition();
    for(int i=0; i!=count; i++) {
      if (vds[i]->val->type == Link) {
        vds[i]->drawIt(p, *this);
      };
    }
    for(int i=0; i!=count; i++) {
      if (vds[i]->val->type == Node) {
        vds[i]->drawIt(p, *this);
      }
    };
    for(int i=0; i!=context_count; i++) {
      context_vds[i]->drawIt(p, *this);
    };
    return true; // always finished if it is no animation
  }

  // do animation
  for(int i=0; i!=count; i++) {
    if (vds[i]->val->type == Link) {
      vds[i]->drawIt(p, *this);
    };
  }
  for(int i=0; i!=count; i++) {
    if (vds[i]->val->type == Node) {
      val = vds[i]->animateIt(p, *this);
        finished = finished && val;
    }
  };
  for(int i=0; i!=context_count; i++) {
    val = context_vds[i]->animateIt(p, *this);
    finished = finished && val;
  };
  return finished;
}


bool gofaiDataView::removeFromContext(gofaiDataViewPointer &dvp) {
  if (dvp.place != gofaiDataViewPointer::Context) {
    kdDebug() << "Attempt to remove something from the context which is not there" << endl;
    return false;
  }
  delete context_vds[dvp.index];
  for(int i=dvp.index;
      i!=context_count-1;i++) {
    context_vds[i]=context_vds[i+1];
  }
  //cerr << "Deleted from there!\n";
  context_count--;
  return true;
}

/*
  save the current context
 */
bool gofaiDataView::writeToDOM(QDomDocument &doc, QDomElement &parent) {
  kdDebug() << "gofaiDataView::writeToDOM\n";
  // flush the data source first
  if (!ds->writeToDOM(doc, parent)) {
      return false;
  }
  QDomElement dataview = doc.createElement("dataview");
  parent.appendChild(dataview);
  dataview.setAttribute("center", center->val->id);
  dataview.setAttribute("context_no", context_count);
  return true;
  /*
  for(int i=0; i < context_count-1;i++) {
    QString buffer;
    buffer.sprintf("context_item_%d",i);
    e.setAttribute(buffer, (Q_UINT32)context_vds[i]->val->id);
  }
  */
}

/*
  load the context
  normally, the recreation stuff should not come here.
*/
bool gofaiDataView::readFromDOM(QDomElement &e) {
  kdDebug() << "gofaiDataView::readFromDOM\n";
  QDomNode n = e;
  while(!n.isNull()) {
      QDomElement e = n.toElement();
      if (e.nodeName() == "items") {
          if (!ds->readFromDOM(e)) {
              kdDebug() << "Error while reading the items!\n";
              return false;
          };
      } else if (e.nodeName() == "dataview") {
          // we assume there is only one...
          gofaiID id =  e.attribute("center");
          gofaiItem *it = ds->lookupItem(id);
          kdDebug() << "after lookup item\n";
          if (it!=NULL) {
              kdDebug() << "Center successfully found\n";
              // fixme: hack
              getView();
              center->val = it; // a bit of a hack
              setCenter(center);
          } else {
              kdDebug() << "No center found set to null\n";
              center = NULL;
          }
          return true; // this is where we exit
      }
      n = n.nextSibling();
  }
  return false;

  /*
  int cnt = e.attribute("context_no").toInt();
  // context_count = cnt;
  for(int i=0; i!=cnt; i++) {
    long id = e.attribute("context_item_%d").toLong();
    // context_vds[i]= ds.lookup(id)
  }
  */
}


